/*
 Stephan Kottler
 kottlers@informatik.uni-tuebingen.de
 *
 */

#include "Settings.hpp"

Settings::Settings():

	var_decay			 	( 135 , 1 ), //= 135/128
	var_min_inc				(1, var_act_t::minNonZeroNumber().getLogOfDenominator() - (2* var_decay.getLogOfDenominator() )),
	var_max_act				(1000),
	cat_decay			 	( 135 , 1 ), //= 135/128
	cat_min_inc				(1, cat_act_t::minNonZeroNumber().getLogOfDenominator() - (2* cat_decay.getLogOfDenominator() )),
	cat_max_act				(1000),
	cls_decay				( 129 , 1 ), //= 129/128
	cls_min_inc				(1, cls_act_t::minNonZeroNumber().getLogOfDenominator() - (2* cls_decay.getLogOfDenominator() )),
	cls_max_act				(1000),

	conflict_max_init		( 100 ),
	conflict_max_inc		( 1.5 ),
	learntsize_factor		( (double)1.0 / (double)3.0),
	learntsize_inc			( 1.05 ),
	solver_name				( NULL )
	//console_out				( NULL )
{
	Option<bool>::val_type val_true	= std::make_pair<const char*,bool>("yes",true);
	Option<bool>::val_type val_false	= std::make_pair<const char*,bool>("no",false);
	//Option<bool> b_opt;

	/*b_opt = Option<bool>(&print_console,true,  "-print-console=",  "If yes, some information is print to console output");
	b_opt.addPossibleValue(val_true);
	b_opt.addPossibleValue(val_false);
	bool_options.push_back(b_opt);*/

	/*b_opt = Option<bool>(&only_bump_on_card_match,	false,  "-only-bump-card=",	"If yes, variable is only bumped if polarity matches");
	b_opt.addPossibleValue(val_true);
	b_opt.addPossibleValue(val_false);
	bool_options.push_back(b_opt);*/
	only_bump_on_card_match = false;

	/*b_opt = Option<bool>(&bump_bin_vars,	false,  "-bump-bins=",		"If yes, variable is bumped each time it is contained in an initial binary clause");
	b_opt.addPossibleValue(val_true);
	b_opt.addPossibleValue(val_false);
	bool_options.push_back(b_opt);*/

	/*b_opt = Option<bool>(&bump_two_left_in_tern,false,  "-bump-two-in-tern=",	"If yes, two free variables in a ternary clause are bumped");
	b_opt.addPossibleValue(val_true);
	b_opt.addPossibleValue(val_false);
	bool_options.push_back(b_opt);*/

	/*b_opt = Option<bool>(&preproc_shuffle,	false,  "-preproc-shuffle=",	"If yes, Vars and Cats are shuffled in preprocess");
	b_opt.addPossibleValue(val_true);
	b_opt.addPossibleValue(val_false);
	bool_options.push_back(b_opt);*/
	preproc_shuffle = true;

	/*b_opt = Option<bool>(&big_simp_at_start,false, "-big-simp=",	"If yes, big simplify is executed once at start");
	b_opt.addPossibleValue(val_true);
	b_opt.addPossibleValue(val_false);
	bool_options.push_back(b_opt);*/

	/*b_opt = Option<bool>(&flip_first_in_preproc,false, "-flip-first-preproc=",	"If yes, first decision in preprocess is always flipped");
	b_opt.addPossibleValue(val_true);
	b_opt.addPossibleValue(val_false);
	bool_options.push_back(b_opt);*/

	/*b_opt = Option<bool>(&sort_watchers,false, "-sort-watchers=",	"If yes, watched lists are sorted by type");
	b_opt.addPossibleValue(val_true);
	b_opt.addPossibleValue(val_false);
	bool_options.push_back(b_opt);*/

	/*b_opt = Option<bool>(&minimise_by_cat,true, "-minimise-cat=",	"If yes, minimisation by category is performed");
	b_opt.addPossibleValue(val_true);
	b_opt.addPossibleValue(val_false);
	bool_options.push_back(b_opt);*/
	minimise_by_cat = true;

	Option<char*>
	s_opt = Option<char*>(&tgf_file_name,NULL, "-tgf=",				"The SAT instance is created from this TGF file (any argument without leading minus is taken as tgf by default). Vertex IDs are assumed to be non-zero positive integers");
	string_options.push_back(s_opt);

	/*s_opt = Option<char*>(&csv_file_name,   NULL,  "-csv-out=",		"If specified, statistics are appended to this file");
	string_options.push_back(s_opt);*/

	s_opt = Option<char*>(&dimacs_out,NULL, 	"-dimacs-out=",		"Encodes the graph-problem into a sat problem without solving it");
	string_options.push_back(s_opt);

	s_opt = Option<char*>(&encoding_info_out, NULL, "-encoding-out=",	"If specified, description of variables is printed to this file");
	string_options.push_back(s_opt);

	Option<uint>
	i_opt = Option<uint>(&rand_seed,	 91648253,  "-seed=",		  "Specify random seed");
	int_options.push_back(i_opt);

	i_opt = Option<uint>(&max_literals,200000000,	"-max-lits=",	  "Maximal number of literals (to restrict mem usage - SAT internal)");
	int_options.push_back(i_opt);

	/*i_opt = Option<uint>(&random_var_freq_permil,20,"-rand-var-freq=", "Random variable frequence [in \%\%]");
	int_options.push_back(i_opt);*/
	random_var_freq_permil = 35;

	/*i_opt = Option<uint>(&random_cat_freq_permil,10,"-rand-cat-freq=", "Random category frequence [in \%\%]");
	int_options.push_back(i_opt);*/
	random_cat_freq_permil = 5;

	/*i_opt = Option<uint>(&random_arc_check_freq_permil,2,"-rand-arc-check=", "Random arc consistency check frequence [in \%\%]");
	int_options.push_back(i_opt);*/
	random_arc_check_freq_permil = 2;

	i_opt = Option<uint>(&timeout,		  ~0,	"-timeout=",		"The solver terminates after the given number of seconds");
	int_options.push_back(i_opt);

	/*i_opt = Option<uint>(&preprocess_level, 1,	"-preprocess=",		"Level of preprocessing [0-2] = non to full");
	int_options.push_back(i_opt);*/
	preprocess_level = 1;

	/*i_opt = Option<uint>(&preprocess_conflicts, 1000,"-preproc-conflicts=",		"Number of conflicts per restart in preprocessing");
	int_options.push_back(i_opt);*/
	preprocess_conflicts = 1000;

	/*i_opt = Option<uint>(&max_tern_closure, 	0,	"-max-tern-closure=","Max number of ternary clauses created by closure");
	int_options.push_back(i_opt);*/
	max_tern_closure = 0;

	/*i_opt = Option<uint>(&arc_consistenct_check, 15,"-arc-const-check=","Perform arc consistency check above stated decision Level");
	int_options.push_back(i_opt);*/
	arc_consistenct_check = 15;


	i_opt = Option<uint>(&grid_width, 	0,	"-grid-width=","Width of grid to use");
	int_options.push_back(i_opt);

	i_opt = Option<uint>(&grid_height, 	0,	"-grid-height=","Height of grid to use");
	int_options.push_back(i_opt);

	/*i_opt = Option<uint>(&bin_cls_to_force_closure, 5000,"-bin-closure-force=","Number of new bin clauses to force closure computation");
	int_options.push_back(i_opt);*/
	bin_cls_to_force_closure = 10000;

	/*i_opt = Option<uint>(&restart_reset, 25,"-restart-reset=","Number of restarts until the conflict limit is reinitialised");
	int_options.push_back(i_opt);*/
	restart_reset = 23;



	/*i_opt = Option<uint>(&polarity_mode, Polarity_Card, "-polarity-mode=","To determine the polarity that is used to assign a decision variable");
	i_opt.addPossibleValue(std::make_pair<const char*,int>("false", Polarity_False  ));
	i_opt.addPossibleValue(std::make_pair<const char*,int>("true",  Polarity_True   ));
	i_opt.addPossibleValue(std::make_pair<const char*,int>("rand",  Polarity_Rand   ));
	i_opt.addPossibleValue(std::make_pair<const char*,int>("prev",  Polarity_Prev   ));
	i_opt.addPossibleValue(std::make_pair<const char*,int>("best",  Polarity_Best   ));
	i_opt.addPossibleValue(std::make_pair<const char*,int>("card",  Polarity_Card   ));
	int_options.push_back(i_opt);*/
	polarity_mode = Polarity_Card;

	/*i_opt = Option<uint>(&encoding_type, Edge, "-encoding=","State which kind of encoding to use");
	i_opt.addPossibleValue(std::make_pair<const char*,int>("direct", Direct));
	i_opt.addPossibleValue(std::make_pair<const char*,int>("edge",  Edge  ));
	i_opt.addPossibleValue(std::make_pair<const char*,int>("relative",  Relative ));
	int_options.push_back(i_opt);*/
}

Settings::~Settings(){}


bool Settings::read(int argc, char** argv){

	//check for correct usage
	solver_name = basename(argv[0]);
	if(argc < 2){
		print_out("Not enough Parameters! \n");
		return false;
	}

	int index(0);
	char* value;
	command_line	= std::string(argv[index]);
	while(++index < argc){

		command_line	+= std::string(argv[index]);

		for(uint opt_idx = 0; opt_idx != bool_options.size();++opt_idx){
			if((value = bool_options[opt_idx].canHandle(argv[index]))){
				if(bool_options[opt_idx].findValue(value))
					goto OptionHandled; else return false;}
		}

		for(uint opt_idx = 0; opt_idx != int_options.size();++opt_idx){
			if((value = int_options[opt_idx].canHandle(argv[index]))){
				if(!int_options[opt_idx].findValue(value)){
					int_options[opt_idx].assign(atoi(value));}
				goto OptionHandled;
			}
		}

		for(uint opt_idx = 0; opt_idx != string_options.size();++opt_idx){
			if((value = string_options[opt_idx].canHandle(argv[index]))){
				string_options[opt_idx].assign(value);
				goto OptionHandled;}
		}

		if(*argv[index] == '-'){
			print_out("Unknown option: %s \n",argv[index]);
			return false;
		}else if(tgf_file_name != NULL){
			print_out("One of the two arguments is incorrect: \"%s\" or \"%s\" \n",tgf_file_name,argv[index]);
			return false;
		}else {
			tgf_file_name	= argv[index];
		}

		OptionHandled: ;
	}

	randy.setSeed(rand_seed);

	return tgf_file_name != NULL;
};


void Settings::printUsage(){

	print_out("Program Usage: \n");
	print_out("%s [Options] \n",solver_name);
	print_out("Options are: \n");
	for(uint opt_idx = 0; opt_idx != bool_options.size();++opt_idx){	bool_options[opt_idx].printDescription();}
	for(uint opt_idx = 0; opt_idx != string_options.size();++opt_idx){  string_options[opt_idx].printDescription();}
	for(uint opt_idx = 0; opt_idx != int_options.size();++opt_idx){	int_options[opt_idx].printDescription();}
	print_out("\n");

	print_out("Current Values are: \n");
	for(uint opt_idx = 0; opt_idx != bool_options.size();++opt_idx){	bool_options[opt_idx].printValue("%i");}
	for(uint opt_idx = 0; opt_idx != string_options.size();++opt_idx){  string_options[opt_idx].printValue("%s");}
	for(uint opt_idx = 0; opt_idx != int_options.size();++opt_idx){	int_options[opt_idx].printValue("%u");}
}


const char*  settings_cpp_version(){return "Settings.cpp:    $Revision: 1.11 $";}
