/*
 Stephan Kottler
 kottlers@informatik.uni-tuebingen.de
 *
 */

#ifndef BASICSOLVER_HPP_
#define BASICSOLVER_HPP_

#include	<vector>

#include	"ConstraintDB.hpp"
#include	"Types.hpp"
#include	"Settings.hpp"
#include	"Statistics.hpp"
#include	"CatActHeap.hpp"
#include	"BitVector.hpp"




using		std::vector;

//BACKJUMP_SET_PREV_RAND,
enum BackjumpType{BACKJUMP_SET_PREV,BACKJUMP_SET_PREV_FLP,BACKJUMP_PREV_UNTOUCHED};




/*
struct LitActDescendingFunctor{
	ActivityMap<Variable,var_act_t>* comp ;
	LitActDescendingFunctor(ActivityMap<Variable,var_act_t>* c):comp(c){};
	bool operator() (const Literal& a, const Literal& b){return comp->less(b.toVar(),a.toVar());}
};

struct LitActAscendingFunctor{
	ActivityMap<Variable,var_act_t>* comp ;
	LitActAscendingFunctor(ActivityMap<Variable,var_act_t>* c):comp(c){};
	bool operator() (const Literal& a, const Literal& b){return comp->less(a.toVar(),b.toVar());}
};
*/

/**
 * Class to decide satisfiablility of a boolean formula in cnf.
 */
class BasicSolver: public ConstraintDB{

	static const int32_t BCP_SUCCESS = -1; //SOMETHING < 0

	/** A heuristic measurement of the activity of a variable  */
	CatActHeap							act_hp;
	vector<int16_t>						tmp_cats_count;
	uint numberCats()					{return tmp_cats_count.size();}

	/** Indicates if state of the solver is ok*/
	bool								state_ok;

	/** Assignment stack; stores all assigments made in the order they were made*/
	vector<Literal>				 		assign_stack;

	/** Head of queue - as index into the assign_stack */
	uint32_t							queue_begin;

	/** Indices for different decision levels in 'assign_stack' */
	vector<uint32_t>					decision_level_end;

	/** Number of top-level assignments at last execution of 'simplify()' */
	uint32_t							simplify_assign_idx;

	/** Remaining number of propagations that must be made before next execution of 'simplify()' */
	int64_t						 		simplify_props_remaining;

	//uint32_t							clause_stamp;

	//current settings
	double  							curr_max_conflicts;
	double  							curr_max_learnts;
	double								curr_rand_var_freq;
	double								curr_rand_cat_freq;
	bool								reduced_learnts;

	//conflict related information
	vector<Literal>						conflict_lemma;
	uint64_t							conflict_involved_levels;
	vector<Variable>					var_flags_to_clear;
	vector<Literal>				 		tmp_literal_store;
	uint								learnts_checked_to;



 //Methods to simplify the clause database
	void 			clearShortWatchers			(Variable var);
	void 			clearShortWatchers			(Literal lit);
	void			simplifyAllShortWatchers 	();
	void			simplifyShortWatchersOf 	(vector<WatchedElement>& wes);
	TrBool		  	simplify					( );

	bool			clsSubsumedByShort(Clause& cl);
	uint			clsFindSubsumed(Clause& cl);
	uint			clsFindSubsumedCardConstraint(Clause& cl);
	TrBool			clsComputeClosure(const Literal& lit0,const we_vec&  watchers, const bool tern );
	TrBool			clsComputeClosureLit(const Literal& lit);
	TrBool			clsComputeBinTernClosure();


	TrBool			checkSubsumed(vector<Clause*>& clses, uint start = 0);
	TrBool 			removeSatAndShrinkClauses	(vector<Clause*>& clses,uint32_t max_cls_sz);
	uint32_t 		clsCheckSatAndShrink		(Clause& cl);
	uint32_t 		clsCheckSatAndShrinkConsAct	(Clause& cl);
	//TrBool 		bigSimplify					( );
	TrBool			preProcess					();
	TrBool			initWithBest				();

//unit propagation
	bool			firstDecisionCheckAll(Category cat);
	bool			firstDecisionCheckAll();
	void			firstDecisionCheckBothPol(Literal next);
	int32_t			bcp		   			(const bool init_conflict );
	int32_t			bcpPropagateLit		(const bool init_conflict, const Literal& false_lit);
	inline void 	bcpFinishOnConflict	(we_vec&  watched, we_iter iter, we_iter copy,we_iter end);
	inline int32_t	bcpInitConflict		(const int32_t& dec_level, const Literal& w1, const Literal& w2);
	inline bool		bcpFindNewWatcher	(Clause& cl, const Literal& false_lit, const uint32_t& other_lit_idx);

//learning
	inline void		backjumpKeepLevel   (uint32_t level, BackjumpType type);
	void			backjumpUndoAssigns (uint32_t assign_idx, BackjumpType type); //undoes all assignments above and including the specified assign_idx
	bool			firstUIP			(int32_t nbr_pivot_lits);
	inline void		firstUIPCheckLiteral(const Literal& lit,const int32_t& dec_level,int32_t& nbr_pivot_elements);

	uint32_t		backtrackLevel		();
	void			reduceSelfImpl		();
	void			tryMakeShortByCardinality ( Clause& cl);
	bool			debugTryMakeShortPossible(Clause& cl, Literal lit0, Literal lit1 = INVALID_LIT, Literal l2= INVALID_LIT);
	inline void		unsetConfLemmaFlags	();
	bool			litIsNecessary	  	(Literal lit);
	inline bool		reduceSelfImpCheckLiteral(const Literal& lit);

	bool			clsImpliedByCardConstraint(const Literal lit0, const Literal lit1);
	inline void		clsLearnUnit(const Literal lit);
	inline bool		clsBinRedundant(const Literal lit0, const Literal lit1);
	inline bool		clsTernRedundant(const Literal lit0, const Literal lit1, const Literal lit2 );


//heuristic
	//inline void		varBump				(const Variable& var);
	inline void		litBump				(const Literal& lit);
	inline void		claDecayActivity	();


//solving
	inline void		reinsertVarOrder	(Variable v);
	inline void		newDecisionLevel	();
	template<class Reason>
	inline void		uncheckedEnqueue	(const Literal& lit, Reason from = (Clause*)NULL);
	inline void		uncheckedEnqueue	(const Literal& lit, const Literal& r1,const Literal& r2 );

	Variable		nextDecisionVar		(const Category& cat);
	Variable		nextNoCatDecision	();
	Literal			nextDecision		(PolarityMode mode);
	Category		nextDecisionCat		();
	Literal			decisionLitPolarity(Variable var, PolarityMode mode,Category cat = Category());
	TrBool			noicySearching		(uint count = 500);
	TrBool			search			  	();
	bool			addValidClause  	(vector<Literal>::iterator lbeg, const vector<Literal>::iterator lend, bool redundant);
	bool			addValidCardConstraint(vector<Literal>::iterator lbeg, const vector<Literal>::iterator lend);

//arc consistency
	TrBool			checkCardConstraints();
	int32_t			bcpCardConstraints();
	int32_t			bcpCardConstPropagateLit(const Literal& false_lit);


public:
					BasicSolver(Settings& sets, Statistics& stats,OFStream* out);
					~BasicSolver();
	inline void		setInvalid();
	inline void		reserveForVariables(uint32_t sz);
			bool	addClause(vector<Literal>& lits, bool redundant = false);
			bool	addCardinalityConstraint(vector<Literal>& lits);
			TrBool  isSAT(double timeout = -1);
			void	printDIMACS(OFStream* out, OFStream* learnts_out);
			void 	printClause(OFStream& out, Clause& cls);
	inline	void	addBestPartialAssign(Literal lit);
	inline uint32_t numberAssignments()  const;
	inline uint32_t decisionLevel	()  const;

};



//Inline Implementation
//>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
void	BasicSolver::setInvalid()	{state_ok = false;}

void	BasicSolver::reinsertVarOrder(Variable v){
	act_hp.restoreVarAndCat(v);
}


void		BasicSolver::litBump(const Literal& lit){
	if(!act_hp.bump(lit,settings.only_bump_on_card_match)) ++statistics.nbr_expensive_var_bmp;
}



template<class Reason>
void	BasicSolver::uncheckedEnqueue(const Literal& lit, Reason from){
	assert(litIsUnknown(lit));
	varUncheckedAssign(lit,from,decisionLevel());
	assign_stack.push_back(lit);
}

void	BasicSolver::uncheckedEnqueue(const Literal& lit, const Literal& r1,const Literal& r2 ){
	assert(litIsUnknown(lit));
	assert(litIsFalse(r1) && litIsFalse(r2));
	varUncheckedAssign(lit,r1,r2,decisionLevel());
	assign_stack.push_back(lit);
}


void		BasicSolver::claDecayActivity()				{
#ifdef CLAUSE_ACT
	cls_inc *= settings.cls_decay;
#endif
}

uint32_t	BasicSolver::decisionLevel ()		const	{return decision_level_end.size(); }

uint32_t	BasicSolver::numberAssignments()	const	{return assign_stack.size();}

void		BasicSolver::newDecisionLevel()				{decision_level_end.push_back(assign_stack.size()); }



void	 	BasicSolver::reserveForVariables(uint32_t sz){
	if(sz > numberVariables()){
		ConstraintDB::reserveForVariables(sz);

		act_hp.reserveForVariables(sz);
	}
}

void BasicSolver::clsLearnUnit(const Literal lit){
	assert(litIsUnknown(lit) && decisionLevel() == 0);
	uncheckedEnqueue<Clause*>(lit);
}

bool BasicSolver::clsBinRedundant(const Literal lit0, const Literal lit1){
return
	clsImpliedByCardConstraint(lit0,lit1) ||
	clsFindBinary(lit0,lit1);
}


bool BasicSolver::clsTernRedundant(const Literal lit0, const Literal lit1, const Literal lit2 ){
return
	clsImpliedByCardConstraint(lit0,lit1) 	||
	clsImpliedByCardConstraint(lit0,lit2)	||
	clsImpliedByCardConstraint(lit1,lit2)	||
	clsFindTernary(lit0,lit1,lit2);
}


void		BasicSolver::backjumpKeepLevel(uint32_t level,BackjumpType type){

	if(decisionLevel() <= level) return;
	backjumpUndoAssigns(decision_level_end[level],type);
	decision_level_end.resize(level);
}


void BasicSolver::bcpFinishOnConflict(we_vec&  watched, we_iter iter, we_iter copy,we_iter end){
	queue_begin 	= assign_stack.size(); // to finish outer while loop
	while(iter != end) *copy++ = *iter++;
	if(copy != end) watched.erase(copy,end);
}

void	BasicSolver::addBestPartialAssign(Literal lit){
	litSetAsPrev(lit);
	act_hp.bump(lit,settings.only_bump_on_card_match);
	++statistics.max_assigned;
}

int32_t	BasicSolver::bcpInitConflict(const int32_t& dec_level, const Literal& w1, const Literal& w2){
	assert(var_flags.allCleared());

	conflict_lemma					.resize(1,INVALID_LIT);
	conflict_involved_levels 		= 0;
	int32_t nbr_pivot_lits 			= 0;
	firstUIPCheckLiteral(w1,dec_level,nbr_pivot_lits);
	firstUIPCheckLiteral(w2,dec_level,nbr_pivot_lits);
	return nbr_pivot_lits;
}


bool	BasicSolver::bcpFindNewWatcher(Clause& cl, const Literal& false_lit, const uint32_t& other_lit_idx){
	uint32_t this_lit_idx = other_lit_idx^1;

	//for(uint32_t k=cl.getNextK(clause_stamp);k<cl.size();++k){
	for(uint32_t k=2;k<cl.size();++k){
		if(litIsNotFalse(cl[k])){
			cl[this_lit_idx] = cl[k];
			cl[k]			= false_lit;
			//cl.setNextK(k+1,clause_stamp);

			//we is "deleted" (not copied) from the watchers list
			litAddWatchedClause(cl[this_lit_idx],cl[other_lit_idx],&cl);
			return true;
		}
	}
	return false;
}


void BasicSolver::firstUIPCheckLiteral(const Literal& lit,const int32_t& dec_level,int& nbr_pivot_elements){
   const Variable	  var(lit.toVar());

   if(var_flags.isSet(var)) return;
   assert(litIsFalse(lit));

   int level = varAssignLevel(var);
   if(level > 0){
	   var_flags			   .setFlag(var);
	   litBump(lit);

	   if(level <  dec_level){
		   conflict_lemma			.push_back(lit);
		   conflict_involved_levels |= hash64(level);
	   }else {
		   ++nbr_pivot_elements;
	   }
   }
}

void BasicSolver::unsetConfLemmaFlags(){
	uint32_t cp = conflict_lemma.size();
	while(cp--) {var_flags.carelessClear(conflict_lemma[cp].toVar());}
	for(cp = var_flags_to_clear.size();cp--;) {var_flags.carelessClear(var_flags_to_clear[cp]);}
	var_flags_to_clear.clear();
	assert(var_flags.allCleared());
}


bool BasicSolver::reduceSelfImpCheckLiteral(const Literal& lit){

	Variable var = lit.toVar();

	if(var_flags.isSet(var)) return false;

	const AssignInfo& a_info = varAssignInfo(var);
	if(a_info.getLevel() > 0){
		if(a_info.noReason() || ((hash64(a_info.getLevel()) & conflict_involved_levels) == 0)){
			return true;
		}

		var_flags.setFlag(var);
		tmp_literal_store.push_back(lit);
	}
	return false;
}



#define for_all_but_one(idx,sz,todo) {todo}; for(idx = 2;idx < sz;++idx){ todo}


const char*  basic_solver_cpp_version();
const char*  filtered_solver_cpp_version();
const char*  sapperlot_cpp_version();

#define basic_solver_hpp_version "BasicSolver.hpp:  $Revision: 1.3 $"


























#endif /*BASICSOLVER_HPP_*/
