/*
 * Algebra.hpp
 *
 *  Created on: 20.07.2009
 *      Author: stephan
 */

#ifndef ALGEBRA_HPP_
#define ALGEBRA_HPP_



inline uint ggT(uint a, uint b){
	assert(a != 0 || b != 0);
	uint rest;
	do {
		rest 	= b % a;
		b 		= a;
		a 		= rest;
	} while (rest != 0);
	return b;
}


struct Point{
	int x;
	int y;
	Point(int a, int b):x(a),y(b){}
	Point& operator*=(int s){
		x*=s; y*=s;	return *this;}
};

inline Point operator-(const Point& a, const Point& b){
	return Point(a.x -b.x, a.y -b.y);
}

inline bool fractionInIntervallZeroOne(int nom, int denom){
//	true iff nom/denom >=0 && <= 1

	assert(denom != 0);
	if(nom == 0) return true;

	return (nom > 0 == denom > 0) ?	abs(nom) <= abs(denom) : false;
}

inline bool scalarInIntervallZeroOne(int left, int right){

	return (left == 0) ? right == 0 : fractionInIntervallZeroOne(right, left);
}


/*
bool consistentLine(int left1, int left2, right){

	if(left1 ==0 ){ return scalarInIntervallZeroOne(left2,right);}

	if(left2 ==0 ){ return scalarInIntervallZeroOne(left1,right);}

	return scalarInIntervallZeroOne(left1,right) || scalarInIntervallZeroOne(left2,right);

}
*/


inline bool validCondition(double left, double right, bool strict){

	if(left == 0) return right == 0;

	double frac = right / left;

	return strict ? (frac > 0 && frac < 1) : (frac >= 0 && frac <= 1);
}

inline bool validConditionsOneLine(double left1, double left2, double right){
	return 		validCondition(left1,right,true) || validCondition(left1,right-left2,true)
			||	validCondition(left2,right,true) || validCondition(left2,right-left1,true);
}


inline bool validConditionsOneZeroEntry(double matrix[2][3]){
	assert(matrix[0][0] == 0 || matrix[0][1] == 0 || matrix[1][0] == 0 ||matrix[1][1] == 0);

	if(matrix[0][0] == 0 && matrix[0][1]==0){
		return matrix[0][2] != 0 ? false:  validConditionsOneLine(matrix[1][0],matrix[1][1],matrix[1][2]);
	}

	if(matrix[1][0] == 0 && matrix[1][1]==0){
		return matrix[1][2] != 0 ? false:  validConditionsOneLine(matrix[0][0],matrix[0][1],matrix[0][2]);
	}

	uint row0 = 0;
	uint row1 = 1;
	uint col0 = 0;
	uint col1 = 1;

	if(matrix[row1][col0] != 0){
		if		(matrix[0][0] == 0){row0=1;row1=0;}
		else if (matrix[1][1] == 0){col0=1;col1=0;}
		else{assert(matrix[0][1]==0);row0=1;row1=0;col0=1;col1=0;}
	}

	assert(matrix[row1][col0] == 0);

	double frac1 = matrix[row1][2]/ matrix[row1][col1];
	if(frac1 < 0 || frac1 > 1) return false;

	return validCondition(matrix[row0][col0],matrix[row0][2]-matrix[row0][col1]*frac1,frac1==0||frac1==1);
}

inline bool twoLinesCross(Point a1, Point a2, Point b1, Point b2){

	double matrix[2][3];
	matrix[0][0] = a2.x -a1.x;
	matrix[1][0] = a2.y -a1.y;
	matrix[0][1] = b1.x -b2.x;
	matrix[1][1] = b1.y -b2.y;
	matrix[0][2] = b1.x -a1.x;
	matrix[1][2] = b1.y -a1.y;

	if(matrix[0][0] != 0 && matrix[0][1] != 0 && matrix[1][0] != 0  && matrix[1][1] != 0){
		double first_scalar 	= matrix[1][0];
		double second_scalar	= matrix[0][0];
		matrix[1][1] = matrix[0][1]*first_scalar - matrix[1][1]*second_scalar;
		matrix[1][2] = matrix[0][2]*first_scalar - matrix[1][2]*second_scalar;
		matrix[1][0] = 0;
	}
	return validConditionsOneZeroEntry(matrix);
}


inline bool contains(Point a1,Point a2, Point b){

	Point a_dir = a2-a1;
	Point cond	= b-a1;

	return 		(a_dir.x*cond.y == a_dir.y*cond.x)
			&& 	scalarInIntervallZeroOne(a_dir.x,cond.x)
			&& 	scalarInIntervallZeroOne(a_dir.y,cond.y);
}






#endif /* ALGEBRA_HPP_ */
